"""
  This module is intended to be imported by command line tools so they can
  configure the root logger so that other loggers used in other modules can
  inherit the configuration.
"""
import logging
import os
import sys

# iam: 6/30/2017 decided to move to a gllvm style where we can set the level and the output file
_loggingEnvLevel_old = 'WLLVM_OUTPUT'
_loggingEnvLevel_new = 'WLLVM_OUTPUT_LEVEL'

_loggingDestination = 'WLLVM_OUTPUT_FILE'

_validLogLevels = ['ERROR', 'WARNING', 'INFO', 'DEBUG']

def logConfig(name):

    destination = os.getenv(_loggingDestination)

    if destination:
        logging.basicConfig(filename=destination, level=logging.WARNING, format='%(levelname)s:%(message)s')
    else:
        logging.basicConfig(level=logging.WARNING, format='%(levelname)s:%(message)s')

    retval = logging.getLogger(name)

    # ignore old setting
    level = os.getenv(_loggingEnvLevel_new)

    if level:
        level = level.upper()
        if not level in _validLogLevels:
            logging.error('"%s" is not a valid value for %s or %s. Valid values are %s',
                          level, _loggingEnvLevel_old, _loggingEnvLevel_new, _validLogLevels)
            sys.exit(1)
        else:
            retval.setLevel(getattr(logging, level))

    # Adjust the format if debugging
    if retval.getEffectiveLevel() == logging.DEBUG:
        formatter = logging.Formatter('%(levelname)s::%(module)s.%(funcName)s() at %(filename)s:%(lineno)d ::%(message)s')
        for h in logging.getLogger().handlers:
            h.setFormatter(formatter)

    return retval

def loggingConfiguration():
    destination = os.getenv(_loggingDestination)
    level = os.getenv(_loggingEnvLevel_new)
    return (destination, level)


def informUser(msg):
    sys.stderr.write(msg)



from __future__ import absolute_import
from __future__ import print_function


import os
import sys
import tempfile
import hashlib

from shutil import copyfile
from .filetype import FileType
from .popenwrapper import Popen
from .arglistfilter import ArgumentListFilter

from .logconfig import logConfig

# Internal logger
_logger = logConfig(__name__)

def wcompile(mode):
    """ The workhorse, called from wllvm and wllvm++.
    """

    rc = 1

    legible_argstring = ' '.join(list(sys.argv)[1:])

    # for diffing with gclang
    _logger.info('Entering CC [%s]', legible_argstring)

    try:
        cmd = list(sys.argv)
        cmd = cmd[1:]

        builder = getBuilder(cmd, mode)

        af = builder.getBitcodeArglistFilter()

        rc = buildObject(builder)

        # phase one compile failed. no point continuing
        if rc != 0:
            _logger.error('Failed to compile using given arguments: [%s]', legible_argstring)
            return rc

        # no need to generate bitcode (e.g. configure only, assembly, ....)
        (skipit, reason) = af.skipBitcodeGeneration()
        if skipit:
            _logger.debug('No work to do: %s', reason)
            _logger.debug(af.__dict__)
            return rc

        # phase two
        buildAndAttachBitcode(builder, af)

    except Exception as e:
        _logger.warning('%s: exception case: %s', mode, str(e))

    _logger.debug('Calling %s returned %d', list(sys.argv), rc)
    return rc




fullSelfPath = os.path.realpath(__file__)
prefix = os.path.dirname(fullSelfPath)
driverDir = prefix
asDir = os.path.abspath(os.path.join(driverDir, 'dragonegg_as'))


# Environmental variable for path to compiler tools (clang/llvm-link etc..)
llvmCompilerPathEnv = 'LLVM_COMPILER_PATH'

# Environmental variable for cross-compilation target.
binutilsTargetPrefixEnv = 'BINUTILS_TARGET_PREFIX'

# This is the ELF section name inserted into binaries
elfSectionName = '.llvm_bc'

# (Fix: 2016/02/16: __LLVM is now used by MacOS's ld so we changed the segment name to __WLLVM).
#
# These are the MACH_O segment and section name
# The SegmentName was __LLVM. Changed to __WLLVM to avoid clashing
# with a segment that ld now uses (since MacOS X 10.11.3?)
#
darwinSegmentName = '__WLLVM'
darwinSectionName = '__llvm_bc'


# Same as an ArgumentListFilter, but DO NOT change the name of the output filename when
# building the bitcode file so that we don't clobber the object file.
class ClangBitcodeArgumentListFilter(ArgumentListFilter):
    def __init__(self, arglist):
        localCallbacks = {'-o' : (1, ClangBitcodeArgumentListFilter.outputFileCallback)}
        super(ClangBitcodeArgumentListFilter, self).__init__(arglist, exactMatches=localCallbacks)

    def outputFileCallback(self, flag, filename):
        self.outputFilename = filename


def getHashedPathName(path):
    return hashlib.sha256(path.encode('utf-8')).hexdigest() if path else None


def attachBitcodePathToObject(bcPath, outFileName):
    # Don't try to attach a bitcode path to a binary.  Unfortunately
    # that won't work.
    (_, ext) = os.path.splitext(outFileName)
    _logger.debug('attachBitcodePathToObject: %s  ===> %s [ext = %s]', bcPath, outFileName, ext)

    #iam: just object files, right?
    fileType = FileType.getFileType(outFileName)
    if fileType not in (FileType.MACH_OBJECT, FileType.ELF_OBJECT):
    #if fileType not in (FileType.MACH_OBJECT, FileType.MACH_SHARED, FileType.ELF_OBJECT, FileType.ELF_SHARED):
        _logger.warning('Cannot attach bitcode path to "%s of type %s"', outFileName, FileType.getFileTypeString(fileType))
        return

    #iam: this also looks very dodgey; we need a more reliable way to do this:
    #if ext not in ('.o', '.lo', '.os', '.So', '.po'):
    #    _logger.warning('Cannot attach bitcode path to "%s of type %s"', outFileName, FileType.getReadableFileType(outFileName))
    #    return

    # Now just build a temporary text file with the full path to the
    # bitcode file that we'll write into the object file.
    f = tempfile.NamedTemporaryFile(mode='w+b', delete=False)
    absBcPath = os.path.abspath(bcPath)
    f.write(absBcPath.encode())
    f.write('\n'.encode())
    _logger.debug('Wrote "%s" to file "%s"', absBcPath, f.name)

    # Ensure buffers are flushed so that objcopy doesn't read an empty
    # file
    f.flush()
    os.fsync(f.fileno())
    f.close()

    binUtilsTargetPrefix = os.getenv(binutilsTargetPrefixEnv)

    # Now write our bitcode section
    if sys.platform.startswith('darwin'):
        objcopyBin = '{}-{}'.format(binUtilsTargetPrefix, 'ld') if binUtilsTargetPrefix else 'ld'
        objcopyCmd = [objcopyBin, '-r', '-keep_private_externs', outFileName, '-sectcreate', darwinSegmentName, darwinSectionName, f.name, '-o', outFileName]
    else:
        objcopyBin = '{}-{}'.format(binUtilsTargetPrefix, 'objcopy') if binUtilsTargetPrefix else 'objcopy'
        objcopyCmd = [objcopyBin, '--add-section', '{0}={1}'.format(elfSectionName, f.name), outFileName]
    orc = 0

    # loicg: If the environment variable WLLVM_BC_STORE is set, copy the bitcode
    # file to that location, using a hash of the original bitcode path as a name
    storeEnv = os.getenv('WLLVM_BC_STORE')
    if storeEnv:
        hashName = getHashedPathName(absBcPath)
        copyfile(absBcPath, os.path.join(storeEnv, hashName))

    try:
        if os.path.getsize(outFileName) > 0:
            objProc = Popen(objcopyCmd)
            orc = objProc.wait()
    except OSError:
        # configure loves to immediately delete things, causing issues for
        # us here.  Just ignore it
        os.remove(f.name)
        sys.exit(0)

    os.remove(f.name)

    if orc != 0:
        _logger.error('objcopy failed with %s', orc)
        sys.exit(-1)

class BuilderBase(object):
    def __init__(self, cmd, mode, prefixPath=None):
        self.af = None     #memoize the arglist filter
        self.cmd = cmd
        self.mode = mode

        # Used as prefix path for compiler
        if prefixPath:
            self.prefixPath = prefixPath
            # Ensure prefixPath has trailing slash
            if self.prefixPath[-1] != os.path.sep:
                self.prefixPath = self.prefixPath + os.path.sep
            # Check prefix path exists
            if not os.path.exists(self.prefixPath):
                errorMsg = 'Path to compiler "%s" does not exist'
                _logger.error(errorMsg, self.prefixPath)
                raise Exception(errorMsg)

        else:
            self.prefixPath = ''

    def getCommand(self):
        if self.af is not None:
            # need to remove things like "-dead_strip"
            forbidden = self.af.forbiddenArgs
            if forbidden:
                for baddy in forbidden:
                    self.cmd.remove(baddy)
        return self.cmd


class ClangBuilder(BuilderBase):
    def getBitcodeCompiler(self):
        cc = self.getCompiler()
        return cc + ['-emit-llvm']

    def getCompiler(self):
        if self.mode == "wllvm++":
            env, prog = 'LLVM_CXX_NAME', 'clang++'
        elif self.mode == "wllvm":
            env, prog = 'LLVM_CC_NAME', 'clang'
        elif self.mode == "wfortran":
            env, prog = 'LLVM_F77_NAME', 'flang'
        else:
            raise Exception("Unknown mode {0}".format(self.mode))
        return ['{0}{1}'.format(self.prefixPath, os.getenv(env) or prog)]

    def getBitcodeArglistFilter(self):
        if self.af is None:
            self.af = ClangBitcodeArgumentListFilter(self.cmd)
        return self.af

class DragoneggBuilder(BuilderBase):
    def getBitcodeCompiler(self):
        pth = os.getenv('LLVM_DRAGONEGG_PLUGIN')
        cc = self.getCompiler()
        # We use '-B' to tell gcc where to look for an assembler.
        # When we build LLVM bitcode we do not want to use the GNU assembler,
        # instead we want gcc to use our own assembler (see as.py).
        cmd = cc + ['-B', asDir, '-fplugin={0}'.format(pth), '-fplugin-arg-dragonegg-emit-ir']
        _logger.debug(cmd)
        return cmd

    def getCompiler(self):
        pfx = ''
        if os.getenv('LLVM_GCC_PREFIX') is not None:
            pfx = os.getenv('LLVM_GCC_PREFIX')

        if self.mode == "wllvm++":
            mode = 'g++'
        elif self.mode == "wllvm":
            mode = 'gcc'
        elif self.mode == "wfortran":
            mode = 'gfortran'
        else:
            raise Exception("Unknown mode {0}".format(self.mode))
        return ['{0}{1}{2}'.format(self.prefixPath, pfx, mode)]

    def getBitcodeArglistFilter(self):
        if self.af is None:
            self.af = ArgumentListFilter(self.cmd)
        return self.af

def getBuilder(cmd, mode):
    compilerEnv = 'LLVM_COMPILER'
    cstring = os.getenv(compilerEnv)
    pathPrefix = os.getenv(llvmCompilerPathEnv) # Optional

    _logger.debug('WLLVM compiler using %s', cstring)
    if pathPrefix:
        _logger.debug('WLLVM compiler path prefix "%s"', pathPrefix)

    if cstring == 'clang':
        return ClangBuilder(cmd, mode, pathPrefix)
    elif cstring == 'dragonegg':
        return DragoneggBuilder(cmd, mode, pathPrefix)
    elif cstring is None:
        errorMsg = ' No compiler set. Please set environment variable %s'
        _logger.critical(errorMsg, compilerEnv)
        raise Exception(errorMsg)
    else:
        errorMsg = '%s = %s : Invalid compiler type'
        _logger.critical(errorMsg, compilerEnv, str(cstring))
        raise Exception(errorMsg)

def buildObject(builder):
    objCompiler = builder.getCompiler()
    objCompiler.extend(builder.getCommand())
    proc = Popen(objCompiler)
    rc = proc.wait()
    _logger.debug('buildObject rc = %d', rc)
    return rc


# This command does not have the executable with it
def buildAndAttachBitcode(builder, af):

    #iam: when we have multiple input files we'll have to keep track of their object files.
    newObjectFiles = []

    hidden = not af.isCompileOnly

    if  len(af.inputFiles) == 1 and af.isCompileOnly:
        _logger.debug('Compile only case: %s', af.inputFiles[0])
        # iam:
        # we could have
        # "... -c -o foo.o" or even "... -c -o foo.So" which is OK, but we could also have
        # "... -c -o crazy-assed.objectfile" which we wouldn't get right (yet)
        # so we need to be careful with the objFile and bcFile
        # maybe python-magic is in our future ...
        srcFile = af.inputFiles[0]
        (objFile, bcFile) = af.getArtifactNames(srcFile, hidden)
        if af.outputFilename is not None:
            objFile = af.outputFilename
            bcFile = af.getBitcodeFileName()
        buildBitcodeFile(builder, srcFile, bcFile)
        attachBitcodePathToObject(bcFile, objFile)

    else:

        for srcFile in af.inputFiles:
            _logger.debug('Not compile only case: %s', srcFile)
            (objFile, bcFile) = af.getArtifactNames(srcFile, hidden)
            if hidden:
                buildObjectFile(builder, srcFile, objFile)
                newObjectFiles.append(objFile)

            if srcFile.endswith('.bc'):
                _logger.debug('attaching %s to %s', srcFile, objFile)
                attachBitcodePathToObject(srcFile, objFile)
            else:
                _logger.debug('building and attaching %s to %s', bcFile, objFile)
                buildBitcodeFile(builder, srcFile, bcFile)
                attachBitcodePathToObject(bcFile, objFile)


    if not af.isCompileOnly:
        linkFiles(builder, newObjectFiles)

    sys.exit(0)

def linkFiles(builder, objectFiles):
    af = builder.getBitcodeArglistFilter()
    outputFile = af.getOutputFilename()
    cc = builder.getCompiler()
    cc.extend(objectFiles)
    cc.extend(af.objectFiles)
    cc.extend(af.linkArgs)
    cc.extend(['-o', outputFile])
    proc = Popen(cc)
    rc = proc.wait()
    if rc != 0:
        _logger.warning('Failed to link "%s"', str(cc))
        sys.exit(rc)


def buildBitcodeFile(builder, srcFile, bcFile):
    af = builder.getBitcodeArglistFilter()
    bcc = builder.getBitcodeCompiler()
    bcc.extend(af.compileArgs)
    bcc.extend(['-c', srcFile])
    bcc.extend(['-o', bcFile])
    _logger.debug('buildBitcodeFile: %s', bcc)
    proc = Popen(bcc)
    rc = proc.wait()
    if rc != 0:
        _logger.warning('Failed to generate bitcode "%s" for "%s"', bcFile, srcFile)
        sys.exit(rc)

def buildObjectFile(builder, srcFile, objFile):
    af = builder.getBitcodeArglistFilter()
    cc = builder.getCompiler()
    cc.extend(af.compileArgs)
    cc.append(srcFile)
    cc.extend(['-c', '-o', objFile])
    _logger.debug('buildObjectFile: %s', cc)
    proc = Popen(cc)
    rc = proc.wait()
    if rc != 0:
        _logger.warning('Failed to generate object "%s" for "%s"', objFile, srcFile)
        sys.exit(rc)

# bd & iam:
#
# case 1 (compileOnly):
#
# if the -c flag exists then so do all the .o files, and we need to
# locate them and produce and embed the bit code.
#
# locating them is easy:
#   either the .o is in the cmdline and we are in the simple case,
#   or else it was generated according to getObjectFilename
#
# we then produce and attach bitcode for each inputFile in the cmdline
#
#
# case 2 (compile and link)
#
#  af.inputFiles is not empty, and compileOnly is false.
#  in this case the .o's may not exist, we must regenerate
#  them in any case.
#
#
# case 3 (link only)
#
# in this case af.inputFiles is empty and we are done
#
#



# Feeping Creaturism:
#
# this is the all important version number used by pip.
#
#
"""
Version History:

1.0.0    - 8/2/2016 initial birth as a pip package.

1.0.1    - 8/2/2016 the rst gets a make over, and doc strings
           became more pervasive.

1.0.2    - 8/4/2016 dragonegg issues. trying to include a polite 'as' wrapper
           (i.e. not a  console_script called as).

1.0.3    - 8/4/2016 travis build fixes.

1.0.4    - 8/4/2016 travis build fixes, and exception handling fixes.

1.0.5    - 8/4/2016 exit value was upsetting travis.

1.0.6    - 8/9/2016 exit codes preserved; important for configure scripts like musl libc.

1.0.7    - 8/9/2016 logical restructuring; smaller bites.

1.0.8    - 8/9/2016 test and doc subdirectories are no longer included.

1.0.9    - 8/25/2016  Python 3.0 import fixes (Will Dietz)

1.0.10   - 9/26/2016  Apple's otool just gets biggier and buggier.

1.0.11   - 9/27/2016  Improved Apple's otool fix.

1.0.12   - 10/27/2016 Common flag support.

1.0.13   - 11/05/2016  pylint spots a few mistakes.

1.0.14   - 11/10/2016  --coverage flag.

1.0.15   - 11/15/2016  pylintification complete.

1.0.16   - 11/16/2016  ooops musl points out I screwed up the exit codes AGAIN.

1.0.17   - 11/23/2016  delcypher  #16  over at travitch's place.

1.0.18   - 4/11/2017  tentative solution to the -emit-llvm "out of context" experience.

1.0.19   - 4/19/2017  fixed a '-o' issue in extract-bc and added the bitcode store feature.

1.1.0    - 4/21/2017  no new features on the horizon, no new bugs?

1.1.1    - 4/25/2017 bugs introduced by the new fetures have hopefully been eradicated.

1.1.2    - 4/26/2017 encoding issues with hashlib in the python 3 swarm.

1.1.3    - 5/20/2017 fortran support via flang (pull #60 over at travitch's place)

1.1.4    - 7/24/2017 improvements motivated by gllvm and logic.

1.1.5    - 3/14/2018 fixes suggested by Alexander Bakst

1.2.0    - 4/24/2018 fixes suggested by building the Linux kernel and trying to harmonize with gllvm.
           4/28/2018 can handle thin archives, can sort bitcode input to llvm-{ar, link} and manifest via the -s switch.
           5/1/2018 can handle archives correctly (deal with multiple files with the same name in the archive).

1.2.1    - 5/13/2018  -fsanitize=  now recognized as a compile AND link flag (mothers day edition)

1.2.2    - 6/1/2018 lots of minor fixes from building big projects (and their dependencies) like tor

1.2.3    - 4/15/2019 The tax day version. Almost a years worth of tweaks from building large things like the Linux kernel.

1.2.4    - 4/15/2019 The tax day version, II. Testing the twine upload.

1.2.5    - 4/17/2019 Fixing the pip package, hopefully.

"""

wllvm_version = '1.2.5'
wllvm_date = 'April 17 2019'



from __future__ import print_function

import os
import sys
import subprocess as sp
import re
import pprint
import tempfile
import shutil
import argparse
import codecs

from .popenwrapper import Popen

from .compilers import llvmCompilerPathEnv
from .compilers import elfSectionName
from .compilers import darwinSegmentName
from .compilers import darwinSectionName
from .compilers import getHashedPathName

from .filetype import FileType

from .logconfig import logConfig, informUser



_logger = logConfig(__name__)

decode_hex = codecs.getdecoder("hex_codec")

def extraction():
    """ This is the entry point to extract-bc.
    """

    (success, pArgs) = extract_bc_args()

    if not success:
        return 1

    if sys.platform.startswith('freebsd') or  sys.platform.startswith('linux'):
        return process_file_unix(pArgs)
    elif sys.platform.startswith('darwin'):
        return process_file_darwin(pArgs)

    #iam: do we work on anything else?
    _logger.error('Unsupported or unrecognized platform: %s', sys.platform)
    return 1



bitCodeArchiveExtension = 'bca'
moduleExtension = 'bc'

# Environmental variable for cross-compilation target.
binutilsTargetPrefixEnv = 'BINUTILS_TARGET_PREFIX'

def getSectionSizeAndOffset(sectionName, filename):
    """Returns the size and offset of the section, both in bytes.

    Use objdump on the provided binary; parse out the fields
    to find the given section.  Parses the output,and
    extracts thesize and offset of that section (in bytes).
    """

    binUtilsTargetPrefix = os.getenv(binutilsTargetPrefixEnv)
    objdumpBin = '{}-{}'.format(binUtilsTargetPrefix, 'objdump') if binUtilsTargetPrefix else 'objdump'
    objdumpCmd = [objdumpBin, '-h', '-w', filename]
    objdumpProc = Popen(objdumpCmd, stdout=sp.PIPE)

    objdumpOutput = objdumpProc.communicate()[0]
    if objdumpProc.returncode != 0:
        _logger.error('Could not dump %s', filename)
        sys.exit(-1)

    for line in [l.decode('utf-8') for l in objdumpOutput.splitlines()]:
        fields = line.split()
        if len(fields) <= 7:
            continue
        if fields[1] != sectionName:
            continue
        try:
            size = int(fields[2], 16)
            offset = int(fields[5], 16)
            return (size, offset)
        except ValueError:
            continue

    # The needed section could not be found
    _logger.warning('Could not find "%s" ELF section in "%s", so skipping this entry.', sectionName, filename)
    return None

def getSectionContent(size, offset, filename):
    """Reads the entire content of an ELF section into a string."""
    with open(filename, mode='rb') as f:
        f.seek(offset)
        d = ''
        try:
            c = f.read(size)
            d = c.decode('utf-8')
        except UnicodeDecodeError:
            _logger.error('Failed to read section containing:')
            print(c)
            raise
        # The linker pads sections with null bytes; our real data
        # cannot have null bytes because it is just text.  Discard
        # nulls.
        return d.replace('\0', '')


# otool hexdata pattern.
otool_hexdata = re.compile(r'^(?:[0-9a-f]{8,16}\t)?([0-9a-f\s]+)$', re.IGNORECASE)

def extract_section_darwin(inputFile):
    """Extracts the section as a string, the darwin version.

    Uses otool to extract the section, then processes it
    to a usable state.

    """
    retval = None

    otoolCmd = ['otool', '-X', '-s', darwinSegmentName, darwinSectionName, inputFile]
    otoolProc = Popen(otoolCmd, stdout=sp.PIPE)

    otoolOutput = otoolProc.communicate()[0]
    if otoolProc.returncode != 0:
        _logger.error('otool failed on %s', inputFile)
        sys.exit(-1)

    lines = otoolOutput.splitlines()

    try:
        octets = []
        for line in lines[1:]:
            m = otool_hexdata.match(line.decode())
            if not m:
                _logger.debug('otool output:\n\t%s\nDID NOT match expectations.', line)
                continue
            octetline = m.group(1)
            octets.extend(octetline.split())
        retval = decode_hex(''.join(octets))[0].splitlines()
        if not retval:
            _logger.error('%s contained no %s segment', inputFile, darwinSegmentName)
    except Exception as e:
        _logger.error('extract_section_darwin: %s', str(e))
    return retval

def extract_section_linux(inputFile):
    """Extracts the section as a string, the *nix version."""
    val = getSectionSizeAndOffset(elfSectionName, inputFile)
    if val is None:
        return []
    (sectionSize, sectionOffset) = val
    content = getSectionContent(sectionSize, sectionOffset, inputFile)
    contents = content.split('\n')
    if not contents:
        _logger.error('%s contained no %s. section is empty', inputFile, elfSectionName)
    return contents


def getStorePath(bcPath):
    storeEnv = os.getenv('WLLVM_BC_STORE')
    if storeEnv:
        hashName = getHashedPathName(bcPath)
        hashPath = os.path.join(storeEnv, hashName)
        if os.path.isfile(hashPath):
            return hashPath
    return None


def getBitcodePath(bcPath):
    """Tries to resolve the whereabouts of the bitcode.

    First, checks if the given path points to an existing bitcode file.
    If it does not, it tries to look for the bitcode file in the store directory given
    by the environment variable WLLVM_BC_STORE.
    """

    if not bcPath or os.path.isfile(bcPath):
        return bcPath

    storePath = getStorePath(bcPath)
    if storePath:
        return storePath
    return bcPath

def linkFiles(pArgs, fileNames):
    linkCmd = [pArgs.llvmLinker, '-v'] if pArgs.verboseFlag else [pArgs.llvmLinker]

    linkCmd.append('-o={0}'.format(pArgs.outputFile))

    fileNames = map(getBitcodePath, fileNames)
    linkCmd.extend([x for x in fileNames if x != ''])

    try:
        linkProc = Popen(linkCmd)
    except OSError as e:
        if e.errno == 2:
            errorMsg = 'Your llvm-link does not seem to be easy to find.\nEither install it or use the -l llvmLinker option.'
        else:
            errorMsg = 'OS error({0}): {1}'.format(e.errno, e.strerror)
        _logger.error(errorMsg)
        raise Exception(errorMsg)

    else:
        exitCode = linkProc.wait()
        _logger.info('%s returned %s', pArgs.llvmLinker, str(exitCode))
    return exitCode


def archiveFiles(pArgs, fileNames):
    retCode = 0
    # We do not want full paths in the archive so we need to chdir into each
    # bitcode's folder. Handle this by calling llvm-ar once for all bitcode
    # files in the same directory

    # Map of directory names to list of bitcode files in that directory
    dirToBCMap = {}
    for bitCodeFile in fileNames:
        dirName = os.path.dirname(bitCodeFile)
        basename = os.path.basename(bitCodeFile)
        if dirName in dirToBCMap:
            dirToBCMap[dirName].append(basename)
        else:
            dirToBCMap[dirName] = [basename]

    _logger.debug('Built up directory to bitcode file list map:\n%s', pprint.pformat(dirToBCMap))

    for (dirname, bcList) in dirToBCMap.items():
        _logger.debug('Changing directory to "%s"', dirname)
        os.chdir(dirname)
        larCmd = [pArgs.llvmArchiver, 'rs', pArgs.outputFile] + bcList
        larProc = Popen(larCmd)
        retCode = larProc.wait()
        if retCode != 0:
            _logger.error('Failed to execute:\n%s', pprint.pformat(larCmd))
            break

    if retCode == 0:
        informUser('Generated LLVM bitcode archive {0}\n'.format(pArgs.outputFile))
    else:
        _logger.error('Failed to generate LLVM bitcode archive')

    return retCode

def extract_from_thin_archive(inputFile):
    """Extracts the paths from the thin archive.

    """
    retval = None

    arCmd = ['ar', '-t', inputFile]         #iam: check if this might be os dependent
    arProc = Popen(arCmd, stdout=sp.PIPE)

    arOutput = arProc.communicate()[0]
    if arProc.returncode != 0:
        _logger.error('ar failed on %s', inputFile)
    else:
        retval = arOutput.splitlines()
    return retval



def handleExecutable(pArgs):

    fileNames = pArgs.extractor(pArgs.inputFile)

    if not fileNames:
        return 1

    if  pArgs.sortBitcodeFilesFlag:
        fileNames = sorted(fileNames)


    if pArgs.manifestFlag:
        writeManifest('{0}.llvm.manifest'.format(pArgs.inputFile), fileNames)

    if pArgs.outputFile is None:
        pArgs.outputFile = pArgs.inputFile + '.' + moduleExtension

    return linkFiles(pArgs, fileNames)


def handleThinArchive(pArgs):

    objectPaths = extract_from_thin_archive(pArgs.inputFile)

    if not objectPaths:
        return 1

    bcFiles = []
    for p in objectPaths:
        _logger.debug('handleThinArchive: processing %s', p)
        contents = pArgs.extractor(p)
        for c in contents:
            if c:
                _logger.debug('\t including %s', c)
                bcFiles.append(str(c))



    return  buildArchive(pArgs, bcFiles)

#iam: do we want to preserve the order in the archive? if so we need to return both the list and the dict.
def fetchTOC(inputFile):
    toc = {}

    arCmd = ['ar', '-t', inputFile]         #iam: check if this might be os dependent
    arProc = Popen(arCmd, stdout=sp.PIPE)

    arOutput = arProc.communicate()[0]
    if arProc.returncode != 0:
        _logger.error('ar failed on %s', inputFile)
        return toc

    lines = arOutput.splitlines()

    for line in lines:
        if line in toc:
            toc[line] += 1
        else:
            toc[line] = 1

    return toc


def extractFile(archive, filename, instance):
    arCmd = ['ar', 'xN', str(instance), archive, filename]         #iam: check if this might be os dependent
    try:
        arP = Popen(arCmd)
    except Exception as e:
        _logger.error(e)
        return False

    arPE = arP.wait()

    if arPE != 0:
        errorMsg = 'Failed to execute archiver with command {0}'.format(arCmd)
        _logger.error(errorMsg)
        return False

    return True



def handleArchiveDarwin(pArgs):
    originalDir = os.getcwd() # This will be the destination

    pArgs.arCmd.append(pArgs.inputFile)

    # Make temporary directory to extract objects to
    tempDir = ''
    bitCodeFiles = []

    try:


        tempDir = tempfile.mkdtemp(suffix='wllvm')
        os.chdir(tempDir)

        # Extract objects from archive
        try:
            arP = Popen(pArgs.arCmd)
        except OSError as e:
            if e.errno == 2:
                errorMsg = 'Your ar does not seem to be easy to find.\n'
            else:
                errorMsg = 'OS error({0}): {1}'.format(e.errno, e.strerror)
            _logger.error(errorMsg)
            raise Exception(errorMsg)

        arPE = arP.wait()

        if arPE != 0:
            errorMsg = 'Failed to execute archiver with command {0}'.format(pArgs.arCmd)
            _logger.error(errorMsg)
            raise Exception(errorMsg)

        _logger.debug(2)

        # Iterate over objects and examine their bitcode inserts
        for (root, _, files) in os.walk(tempDir):
            _logger.debug('Exploring "%s"', root)
            for f in files:
                fPath = os.path.join(root, f)
                if FileType.getFileType(fPath) == pArgs.fileType:

                    # Extract bitcode locations from object
                    contents = pArgs.extractor(fPath)

                    for bcFile in contents:
                        if bcFile != '':
                            if not os.path.exists(bcFile):
                                _logger.warning('%s lists bitcode library "%s" but it could not be found', f, bcFile)
                            else:
                                bitCodeFiles.append(bcFile)
                else:
                    _logger.info('Ignoring file "%s" in archive', f)

        _logger.info('Found the following bitcode file names to build bitcode archive:\n%s', pprint.pformat(bitCodeFiles))

    finally:
        # Delete the temporary folder
        _logger.debug('Deleting temporary folder "%s"', tempDir)
        shutil.rmtree(tempDir)

    #write the manifest file if asked for
    if pArgs.manifestFlag:
        manifestFile = '{0}.llvm.manifest'.format(pArgs.inputFile)
        with open(manifestFile, 'w') as output:
            for f in bitCodeFiles:
                output.write('{0}\n'.format(f))

    # Build bitcode archive
    os.chdir(originalDir)

    return buildArchive(pArgs, bitCodeFiles)



#iam: 5/1/2018
def handleArchiveLinux(pArgs):
    """ handleArchiveLinux processes a archive, and creates either a bitcode archive, or a module, depending on the flags used.

    Archives on Linux are strange beasts. handleArchive processes the archive by:

      1. first creating a table of contents of the archive, which maps file names (in the archive) to the number of
    times a file with that name is stored in the archive.

      2. for each OCCURENCE of a file (name and count) it extracts the section from the object file, and adds the
    bitcode paths to the bitcode list.

      3. it then either links all these bitcode files together using llvm-link,  or else is creates a bitcode
    archive using llvm-ar

    """

    inputFile = pArgs.inputFile

    originalDir = os.getcwd() # We want to end up back where we started.

    toc = fetchTOC(inputFile)

    if not toc:
        _logger.warning('No files found, so nothing to be done.')
        return 0

    bitCodeFiles = []

    try:
        tempDir = tempfile.mkdtemp(suffix='wllvm')
        os.chdir(tempDir)

        for filename in toc:
            count = toc[filename]
            for i in range(1, count + 1):

                # extact out the ith instance of filename
                if extractFile(inputFile, filename, i):
                    # Extract bitcode locations from object
                    contents = pArgs.extractor(filename)
                    _logger.debug('From instance %s of %s in %s we extracted\n\t%s\n', i, filename, inputFile, contents)
                    if contents:
                        for path in contents:
                            if path:
                                bitCodeFiles.append(path)
                    else:
                        _logger.debug('From instance %s of %s in %s we extracted NOTHING\n', i, filename, inputFile)

    finally:
        # Delete the temporary folder
        _logger.debug('Deleting temporary folder "%s"', tempDir)
        shutil.rmtree(tempDir)

    _logger.debug('From instance %s we extracted\n\t%s\n', inputFile, bitCodeFiles)

    # Build bitcode archive
    os.chdir(originalDir)

    return buildArchive(pArgs, bitCodeFiles)






def buildArchive(pArgs, bitCodeFiles):

    if pArgs.bitcodeModuleFlag:
        _logger.info('Generating LLVM Bitcode module from an archive')
    else:
        _logger.info('Generating LLVM Bitcode archive from an archive')

    if  pArgs.sortBitcodeFilesFlag:
        bitCodeFiles = sorted(bitCodeFiles)

    #write the manifest file if asked for
    if pArgs.manifestFlag:
        writeManifest('{0}.llvm.manifest'.format(pArgs.inputFile), bitCodeFiles)

    if pArgs.bitcodeModuleFlag:

        # Pick output file path if outputFile not set
        if pArgs.outputFile is None:
            pArgs.outputFile = pArgs.inputFile
            pArgs.outputFile += '.' + moduleExtension

        informUser('Writing output to {0}\n'.format(pArgs.outputFile))
        return linkFiles(pArgs, bitCodeFiles)

    else:

        # Pick output file path if outputFile not set
        if pArgs.outputFile is None:
            bcaExtension = '.' + bitCodeArchiveExtension
            if pArgs.inputFile.endswith('.a'):
                # Strip off .a suffix
                pArgs.outputFile = pArgs.inputFile[:-2]
                pArgs.outputFile += bcaExtension
            else:
                pArgs.outputFile = pArgs.inputFile + bcaExtension

        informUser('Writing output to {0}\n'.format(pArgs.outputFile))
        return archiveFiles(pArgs, bitCodeFiles)


def writeManifest(manifestFile, bitCodeFiles):
    with open(manifestFile, 'w') as output:
        for f in bitCodeFiles:
            output.write('{0}\n'.format(f))
            sf = getStorePath(f)
            if sf:
                output.write('{0}\n'.format(sf))
    _logger.warning('Manifest written to %s', manifestFile)



class ExtractedArgs(object):

    def __init__(self):
        self.fileType = None
        self.outputFile = None
        self.inputFile = None
        self.output = None
        self.extractor = None
        self.arCmd = None


def extract_bc_args():

    # do we need a path in front?
    llvmToolPrefix = os.getenv(llvmCompilerPathEnv)
    if not llvmToolPrefix:
        llvmToolPrefix = ''

    # is our linker called something different?
    llvmLinkerName = os.getenv('LLVM_LINK_NAME')
    if not llvmLinkerName:
        llvmLinkerName = 'llvm-link'
    llvmLinker = os.path.join(llvmToolPrefix, llvmLinkerName)

    # is our archiver called something different?
    llvmArchiverName = os.getenv('LLVM_AR_NAME')
    if not llvmArchiverName:
        llvmArchiverName = 'llvm-ar'
    llvmArchiver = os.path.join(llvmToolPrefix, llvmArchiverName)

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(dest='inputFile',
                        help='A binary produced by wllvm/wllvm++')
    parser.add_argument('--linker', '-l',
                        dest='llvmLinker',
                        help='The LLVM bitcode linker to use. Default "%(default)s"',
                        default=llvmLinker)
    parser.add_argument('--archiver', '-a',
                        dest='llvmArchiver',
                        help='The LLVM bitcode archiver to use. Default "%(default)s"',
                        default=llvmArchiver)
    parser.add_argument('--verbose', '-v',
                        dest='verboseFlag',
                        help='Call the external procedures in verbose mode.',
                        action="store_true")
    parser.add_argument('--manifest', '-m',
                        dest='manifestFlag',
                        help='Write a manifest file listing all the .bc files used.',
                        action='store_true')
    parser.add_argument('--sort', '-s',
                        dest='sortBitcodeFilesFlag',
                        help='Sort the list of bitcode files (for debugging).',
                        action='store_true')
    parser.add_argument('--bitcode', '-b',
                        dest='bitcodeModuleFlag',
                        help='Extract a bitcode module rather than an archive. ' +
                        'Only useful when extracting from an archive.',
                        action='store_true')
    parser.add_argument('--output', '-o',
                        dest='outputFile',
                        help='The output file. Defaults to a file in the same directory ' +
                        'as the input with the same name as the input but with an ' +
                        'added file extension (.'+ moduleExtension + ' for bitcode '+
                        'modules and .' + bitCodeArchiveExtension +' for bitcode archives)',
                        default=None)
    pArgs = parser.parse_args(namespace=ExtractedArgs())


    # Check file exists
    if not os.path.exists(pArgs.inputFile):
        _logger.error('File "%s" does not exist.', pArgs.inputFile)
        return (False, None)

    pArgs.inputFile = os.path.abspath(pArgs.inputFile)


    # Check output destitionation if set
    outputFile = pArgs.outputFile
    if outputFile != None:
        # Get Absolute output path
        outputFile = os.path.abspath(outputFile)
        if not os.path.exists(os.path.dirname(outputFile)):
            _logger.error('Output directory "%s" does not exist.', os.path.dirname(outputFile))
            return (False, None)

    pArgs.output = outputFile

    return (True, pArgs)




def process_file_unix(pArgs):
    retval = 1
    ft = FileType.getFileType(pArgs.inputFile)
    _logger.debug('Detected file type is %s', FileType.revMap[ft])

    pArgs.arCmd = ['ar', 'xv'] if pArgs.verboseFlag else ['ar', 'x']
    pArgs.extractor = extract_section_linux
    pArgs.fileType = FileType.ELF_OBJECT

    if ft == FileType.ELF_EXECUTABLE or ft == FileType.ELF_SHARED or ft == FileType.ELF_OBJECT:
        _logger.info('Generating LLVM Bitcode module')
        retval = handleExecutable(pArgs)
    elif ft == FileType.ARCHIVE:
        retval = handleArchiveLinux(pArgs)
    elif ft == FileType.THIN_ARCHIVE:
        retval = handleThinArchive(pArgs)
    else:
        _logger.error('File "%s" of type %s cannot be used', pArgs.inputFile, FileType.revMap[ft])
    return retval



def process_file_darwin(pArgs):
    retval = 1
    ft = FileType.getFileType(pArgs.inputFile)
    _logger.debug('Detected file type is %s', FileType.revMap[ft])

    pArgs.arCmd = ['ar', '-x', '-v'] if pArgs.verboseFlag else ['ar', '-x']
    pArgs.extractor = extract_section_darwin
    pArgs.fileType = FileType.MACH_OBJECT

    if ft == FileType.MACH_EXECUTABLE or ft == FileType.MACH_SHARED or ft == FileType.MACH_OBJECT:
        _logger.info('Generating LLVM Bitcode module')
        retval = handleExecutable(pArgs)
    elif ft == FileType.ARCHIVE:
        _logger.info('Handling archive')
        retval = handleArchiveDarwin(pArgs)


    else:
        _logger.error('File "%s" of type %s cannot be used', pArgs.inputFile, FileType.revMap[ft])
    return retval



from setuptools import setup, find_packages
import os
import glob

from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# use the in house version number so we stay in synch with ourselves.
from wllvm.version import wllvm_version

setup(
    name='wllvm',
    version=wllvm_version,
    description='Whole Program LLVM',
    long_description=long_description,
    url='https://github.com/SRI-CSL/whole-program-llvm',
    author='Ian A. Mason, Tristan Ravitch, Dan Liew, Bruno Dutertre, Benjamin Schubert, Berkeley Churchill, Marko Dimjasevic, Will Dietz, Fabian Mager, Ben Liblit, Andrew Santosa, Tomas Kalibera, Loic  Gelle, Joshua Cranmer, Alexander Bakst, Miguel Arroyo.',
    author_email='iam@csl.sri.com',


    include_package_data=True,

    packages=find_packages(),

    entry_points = {
        'console_scripts': [
            'wllvm-as = wllvm.as:main',
            'wllvm = wllvm.wllvm:main',
            'wllvm++ = wllvm.wllvmpp:main',
            'wfortran = wllvm.wfortran:main',
            'wllvm-sanity-checker = wllvm.sanity:main',
            'extract-bc = wllvm.extractor:main',
        ],
    },

    license='MIT',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Compilers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Operating System :: POSIX :: BSD',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ],
)



"""
Module support for the wllvm-sanity-checker tool.

The wllvm-sanity-checker tool examines the users
environment to see if it makes sense from the
wllvm point of view. Useful first step in trying to
debug a failure.
"""
from __future__ import print_function

import sys
import os
import subprocess as sp
import errno

from .version import wllvm_version, wllvm_date
from .logconfig import loggingConfiguration

explain_LLVM_COMPILER = """

The environment variable 'LLVM_COMPILER' is a switch. It should either
be set to 'clang' or 'dragonegg'. Anything else will cause an error.

"""

explain_LLVM_DRAGONEGG_PLUGIN = """

You need to set the environment variable LLVM_DRAGONEGG_PLUGIN to the
full path to your dragonegg plugin. Thanks.

"""

explain_LLVM_CC_NAME = """

If your clang compiler is not called clang, but something else, then
you will need to set the environment variable LLVM_CC_NAME to the
appropriate string. For example if your clang is called clang-3.5 then
LLVM_CC_NAME should be set to clang-3.5.

"""

explain_LLVM_CXX_NAME = """

If your clang++ compiler is not called clang++, but something else,
then you will need to set the environment variable LLVM_CXX_NAME to
the appropriate string. For example if your clang++ is called ++clang
then LLVM_CC_NAME should be set to ++clang.

"""

explain_LLVM_COMPILER_PATH = """

Your compiler should either be in your PATH, or else located where the
environment variable LLVM_COMPILER_PATH indicates. It can also be used
to indicate the directory that contains the other LLVM tools such as
llvm-link, and llvm-ar.

"""

explain_LLVM_LINK_NAME = """

If your llvm linker is not called llvm-link, but something else, then
you will need to set the environment variable LLVM_LINK_NAME to the
appropriate string. For example if your llvm-link is called llvm-link-3.5 then
LLVM_LINK_NAME should be set to llvm-link-3.5.

"""

explain_LLVM_AR_NAME = """

If your llvm archiver is not called llvm-ar, but something else,
then you will need to set the environment variable LLVM_AR_NAME to
the appropriate string. For example if your llvm-ar is called llvm-ar-3.5
then LLVM_AR_NAME should be set to llvm-ar-3.5.

"""

class Checker(object):
    def __init__(self):
        path = os.getenv('LLVM_COMPILER_PATH')

        if path and path[-1] != os.path.sep:
            path = path + os.path.sep

        self.path = path if path else ''

    def check(self):
        """Performs the environmental sanity check.

        Performs the following checks in order:
        0. Prints out the logging configuartion
        1. Check that the OS is supported.
        2. Checks that the compiler settings make sense.
        3. Checks that the needed LLVM utilities exists.
        4. Check that the store, if set, exists.
        """

        self.checkSelf()

        self.checkLogging()

        if not self.checkOS():
            print('I do not think we support your OS. Sorry.')
            return 1

        success = self.checkCompiler()

        if success:
            self.checkAuxiliaries()
            self.checkStore()

        return 0 if success else 1

    def checkSelf(self):
        print('wllvm version: {0}'.format(wllvm_version))
        print('wllvm released: {0}\n'.format(wllvm_date))


    def checkLogging(self):
        (destination, level) = loggingConfiguration()
        print('Logging output to {0}.'.format(destination if destination else 'standard error'))
        if not level:
            print('Logging level not set, defaulting to WARNING\n')
        else:
            print('Logging level set to {0}.\n'.format(level))


    def checkOS(self):
        """Returns True if we support the OS."""
        return (sys.platform.startswith('freebsd') or
                sys.platform.startswith('linux') or
                sys.platform.startswith('darwin'))


    def checkSwitch(self):
        """Checks the correctness of the LLVM_COMPILER env var."""
        compiler_type = os.getenv('LLVM_COMPILER')
        if compiler_type == 'clang':
            return (1, '\nWe are using clang.\n')
        elif compiler_type == 'dragonegg':
            return (2, '\nWe are using dragonegg.\n')
        return (0, explain_LLVM_COMPILER)


    def checkClang(self):
        """Checks for clang and clang++."""
        cc_name = os.getenv('LLVM_CC_NAME')
        cxx_name = os.getenv('LLVM_CXX_NAME')

        cc = '{0}{1}'.format(self.path, cc_name if cc_name else 'clang')
        cxx = '{0}{1}'.format(self.path, cxx_name if cxx_name else 'clang++')

        return self.checkCompilers(cc, cxx)


    def checkDragonegg(self):
        """Checks for gcc, g++ and the dragonegg plugin."""
        if not self.checkDragoneggPlugin():
            return False

        pfx = ''
        if os.getenv('LLVM_GCC_PREFIX') is not None:
            pfx = os.getenv('LLVM_GCC_PREFIX')

        cc = '{0}{1}gcc'.format(self.path, pfx)
        cxx = '{0}{1}g++'.format(self.path, pfx)

        return self.checkCompilers(cc, cxx)


    def checkDragoneggPlugin(self):
        """Checks for the dragonegg plugin."""
        plugin = os.getenv('LLVM_DRAGONEGG_PLUGIN')

        if not plugin:
            print(explain_LLVM_DRAGONEGG_PLUGIN)
            return False

        if os.path.isfile(plugin):
            try:
                open(plugin)
            except IOError as e:
                print("Unable to open {0}: {1}".format(plugin, str(e)))
            else:
                return True
        else:
            print("Could not find {0}".format(plugin))
            return False


    def checkCompiler(self):
        """Determines the chosen compiler, and checks it."""
        (code, comment) = self.checkSwitch()

        if code == 0:
            print(comment)
            return False
        elif code == 1:
            print(comment)
            return self.checkClang()
        elif code == 2:
            print(comment)
            return self.checkDragonegg()
        print('Insane')
        return False



    def checkCompilers(self, cc, cxx):
        """Tests that the compilers actually exist."""
        (ccOk, ccVersion) = self.checkExecutable(cc)
        (cxxOk, cxxVersion) = self.checkExecutable(cxx)

        if not ccOk:
            print('The C compiler {0} was not found or not executable.\nBetter not try using wllvm!\n'.format(cc))
        else:
            print('The C compiler {0} is:\n\n\t{1}\n'.format(cc, extractLine(ccVersion, 0)))

        if not cxxOk:
            print('The CXX compiler {0} was not found or not executable.\nBetter not try using wllvm++!\n'.format(cxx))
        else:
            print('The C++ compiler {0} is:\n\n\t{1}\n'.format(cxx, extractLine(cxxVersion, 0)))

        if not ccOk or  not cxxOk:
            print(explain_LLVM_COMPILER_PATH)
            if not ccOk:
                print(explain_LLVM_CC_NAME)
            if not cxxOk:
                print(explain_LLVM_CXX_NAME)



        return ccOk or cxxOk


    def checkExecutable(self, exe, version_switch='-v'):
        """Checks that an executable exists, and is executable."""
        cmd = [exe, version_switch]
        try:
            compiler = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.PIPE)
            output = compiler.communicate()
            compilerOutput = '{0}{1}'.format(output[0], output[1])
        except OSError as e:
            if e.errno == errno.EPERM:
                return (False, '{0} not executable'.format(exe))
            elif e.errno == errno.ENOENT:
                return (False, '{0} not found'.format(exe))
            return (False, '{0} not sure why, errno is {1}'.format(exe, e.errno))
        else:
            return (True, compilerOutput)



    def checkAuxiliaries(self):
        """Checks for the archiver and linker."""
        link_name = os.getenv('LLVM_LINK_NAME')
        ar_name = os.getenv('LLVM_AR_NAME')

        if not link_name:
            link_name = 'llvm-link'

        if not ar_name:
            ar_name = 'llvm-ar'

        link = '{0}{1}'.format(self.path, link_name) if self.path else link_name
        ar = '{0}{1}'.format(self.path, ar_name) if self.path else ar_name

        (linkOk, linkVersion) = self.checkExecutable(link, '-version')

        (arOk, arVersion) = self.checkExecutable(ar, '-version')

        if not linkOk:
            print('The bitcode linker {0} was not found or not executable.\nBetter not try using extract-bc!\n'.format(link))
            print(explain_LLVM_LINK_NAME)
        else:
            print('The bitcode linker {0} is:\n\n\t{1}\n'.format(link, extractLine(linkVersion, 1)))

        if not arOk:
            print('The bitcode archiver {0} was not found or not executable.\nBetter not try using extract-bc!\n'.format(ar))
            print(explain_LLVM_AR_NAME)
        else:
            print('The bitcode archiver {0} is:\n\n\t{1}\n'.format(ar, extractLine(arVersion, 1)))


    def checkStore(self):
        """Checks that the bitcode store, if set, makes sense."""
        store_dir = os.getenv('WLLVM_BC_STORE')
        if store_dir:
            if os.path.exists(store_dir) and os.path.isdir(store_dir) and os.path.isabs(store_dir):
                print('Using the bitcode store:\n\n\t{0}\n\n'.format(store_dir))
            else:
                print('The bitcode store:\n\n\t{0}\n\nis either not absolute, does not exist, or is not a directory.\n\n'.format(store_dir))
        else:
            print('Not using a bitcode store.\n\n')


def extractLine(version, n):
    if not version:
        return version
    lines = version.split('\n')
    line = lines[n] if n < len(lines) else lines[-1]
    return line.strip() if line else line
